#ifndef RAWDATABLOCK_H
#define RAWDATABLOCK_H

#include "DAQDataFormats/RawDataBlockHeader.h"
#include "DAQDataFormats/RawMicroBlock.h"

#include <stdarg.h>		// va_start

namespace daqdataformats {

    #define DEFAULT_DATABLOCK_SIZE 262144 /* Maximum MicroBlock Buffer Size 64k */

    class RawDataBlock : public RawDAQData {
      
    public:
      
      /// Constructors
      RawDataBlock (uint32_t buffersize = DEFAULT_DATABLOCK_SIZE, uint32_t tlvl=0);                        //! The default constructor
      RawDataBlock (const RawDataBlock& copy_in);  //! Deep copy constructor
      RawDataBlock (void* dadablock_buffer, uint32_t buffersize = DEFAULT_DATABLOCK_SIZE);       //! Shallow copy constructor
      /// Destructor
      ~RawDataBlock();

      /// Initialization of the data members
      void init();
      
      /// Deep Copy
      void  copy(const RawDataBlock& copy_in);
      
      /// Reading method to read the whole datablock
      void* readData(void* datablock_buffer);
      // Suplementary reading methods that needs to be defined since readData is overloaded
      void* readData(void* buffer, uint32_t size);
      void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);
      virtual readDataStatus readData( int fd );

      /// Add the MicroBlock to _Buffer.
      /// If check_mc_flag = true, the method checks whether the microslice
      /// in the microblock is MC generatedand adds an appropriate flag to the DataBlock Header.
      void addMicroBlock(void* microblock_buffer, bool check_mc_flag = true);
      /// Add the MicroBlock of a known size to _Buffer.
      /// If check_mc_flag = true, the method checks whether the microslice
      /// in the microblock is MC generatedand adds an appropriate flag to the DataBlock Header.
      void addMicroBlock(void* microblock_buffer, uint32_t microblock_size, bool check_mc_flag = true);
      /// Add the MicroBlock to _Buffer by passing the references of
      /// the MilliSlice and the MicroSlice. MilliSlice is the one that contains the MicroSlice.
      /// If check_mc_flag = true, the method checks whether the microslice
      /// in the microblock is MC generatedand adds an appropriate flag to the DataBlock Header.
      void addMicroBlock(void* millislice_buffer, void* microslice_buffer, bool check_mc_flag = true);
      /// Add the MicroBlock to _Buffer by passing the references of
      /// the MilliSlice and the MicroSlice of a known size.
      /// MilliSlice is the one that contains the MicroSlice.
      /// If check_mc_flag = true, the method checks whether the microslice
      /// in the microblock is MC generatedand adds an appropriate flag to the DataBlock Header.
      void addMicroBlock(void* millislice_buffer, void* microslice_buffer, uint32_t microslice_size, bool check_mc_flag = true);


      /// Method to close the DataBlock. Produces the CRC.
      /// if produce_crc=flase, don't calculate CRC.
      bool close(bool produce_crc = true);



      // Getters
      RawDataBlockHeader* getHeader();
      RawMicroBlock*      getFloatingMicroBlock();
      uint32_t            getVersion()                 const;
      uint32_t            getMarker()                  const;
      uint32_t            getTrigNumLo()               const;
      uint32_t            getTrigNumHi()               const;
      uint64_t            getTrigNum()                 const;
      uint16_t            getNumMicroBlocks()          const;
      uint16_t            getBuffId()                  const;
      uint32_t            getDataSize()                const;
      bool                getMonteCarloFlag()          const;
      bool                isCRCCalculationUsed()       const;
      bool                isConnectionInitialization() const;
      bool                isCloseConnection()          const;
      bool                isMissingData()              const;
      uint32_t            getFreeMemory()              const; //! Free memory left in the buffer.

      //Setters
      bool setVersion                  (uint32_t);             //! DataBlock version number
      bool setTrigNumLo                (uint32_t);             //! Set the trigger number low part
      bool setTrigNumHi                (uint32_t);             //! Set the trigger number high part
      bool setTrigNum                  (uint64_t);             //! Set the trigger number
      bool setBuffId                   (uint16_t);             //! Buffer ID
      bool setFloatingMicroBlock       (uint32_t imicroblock); //! Set Floating MicroBlock
      bool setConnectionInitialization (bool);
      bool setMonteCarloFlag           (bool);
      bool setCloseConnection          (bool);
      bool setMissingData              (bool);


      bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
      bool resetBufferSource();             //!< Reset the class to use its internal buffer
      bool clear();                         //!< Clear the buffer
      bool reset();                         //!< Reset the class data
      bool isInternalBuffer() const;        //!< Is the buffer internal?


      /// Print the DataBlock structure
      void print();

      /// Size of the Data Block in words
      uint32_t sizeofdata() const;

    private:
      
      RawDataBlockHeader   _Header;                  //! DataBlock header
      RawMicroBlock        _FloatingMicroBlock;      //! Floating MicroBlock
      int                  _FloatingMicroBlockNumber;//! Floating MicroBlock Number

      size_t                _fd_read_bytes;
      //! Maximum size of the current buffer
      uint32_t              _max_buffer_size;

      uint32_t              _tlvl; // debug "trace level mask"

      /// The actual internal buffer (not a pointer)
      std::vector<uint32_t> _InternalBuffer;

      void tprintf( unsigned lvl, const char *fmt, ... )
	__attribute__((format(printf, 3, 4))); // g++ note: attributes are not
               			          //  allowed on a function-definition
    };// end of class definition
  
} // END NAMESPCE daqdataformats
#endif /* RAWDATABLOCK_H */
