#ifndef RAWMICROSLICE_H
#define RAWMICROSLICE_H

#include "DAQDataFormats/FEBSimulator.h"
#include "DAQDataFormats/RawMicroSliceHeader.h"
#include "DAQDataFormats/RawTimingMarker.h"
#include "DAQDataFormats/RawNanoSlice.h"

#include <vector>		/* Use the Vector STL */

namespace daqdataformats {

#define DEFAULT_MICROSLICE_SIZE 16384 /* Maximum Microslice Buffer Size 64k/4 */
#define MAXIMUM_FRONTEND_BOARDS 64    /* Maximum number of front end boards we support */

/*
// FLAG Setting for Initialization and Resets
#define MICROSLICE_RESET_MEM 0x1
#define MICROSLICE_RESET_FEB 0x2
#define MICROSLICE_DELETE_FEB 0x4
#define MICROSLICE_INIT_FEB  0x8
*/
  
  class RawMicroSlice : public RawDAQData {
    
  public:
    
    // Constructors
    RawMicroSlice (uint32_t buffersize = DEFAULT_MICROSLICE_SIZE); //! The default constructor
    RawMicroSlice (const RawMicroSlice& copy_in);                  //! Deep copy constructor
    RawMicroSlice (void* microslice_buffer, uint32_t buffersize = DEFAULT_MICROSLICE_SIZE); //! Shallow copy constructor
    /// Destructor
    ~RawMicroSlice();

    /// initialize the MicroSlice
    void init();

    /// Deep Copy
    void  copy(const RawMicroSlice& copy_in);

    /// Add a NanoSlice to the MicroSlice changing the MicroSlice Header appropriately
    /// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
    /// and adds an appropriate flag to the MicroSlice Header.
    void addNanoSlice(void* nanoslice_buffer, bool check_mc_flag = true);
    /// Add a Some nwords of data to the internal buffer changing the MicroSlice Header appropriately
    /// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
    /// and adds an appropriate flag to the MicroSlice Header.
    void addData(void* buffer, uint32_t nwords, bool check_mc_flag = true);

    // Reading methods
    /// This is to read an entire MicroSlice from microslice_buffer into _Buffer
    void* readData(void* microslice_buffer);
    // Suplementary reading methods
    void* readData(void* buffer, uint32_t size);
    void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);

    /// Print out the MicroSlice information
    void print();
    
    /// Generate a MicroSlice
    void generate();

    // Getters
    RawMicroSliceHeader* getHeader();               //! Get the MicroSlice Header
    RawTimingMarker*     getTimingMarker();         //! Get the MicroSlice Timing Marker
    RawNanoSlice*        getFloatingNanoSlice();    //! Get Floating NanoSlice
    bool                 getDataPresent()    const; //! Get the Data Present bit of the Header
    bool                 getMonteCarloFlag() const; //! Get the Data Present bit of the Header
    uint32_t             getByteCount()      const; //! Get the byte count value of the header
    uint32_t             getLowWord()        const; //! Get method for lower half of timing word
    uint32_t             getHighWord()       const; //! Get method for upper half of timing word
    uint64_t             getTime()           const; //! Get the time-stamp as a single 64-bit value
    float                getHitProbability() const; //! Get current Hit Probability (for a microslice generation)
    uint32_t             getNumNanoSlices()  const; //! Get number of NanoSlices in the MicroSlice
    uint32_t             getFreeMemory()     const; //! Free memory left in the buffer.

    //Setters
    bool setMonteCarloFlag    (bool);            //! Set the SIM flag
    bool setFloatingNanoSlice (uint32_t inano);  //! Set Floating NanoSlice to point to the "inano"-th NanoSlice
    bool setLowWord           (uint32_t);	 //! Set method for lower half of timing word
    bool setHighWord          (uint32_t);	 //! Set method for upper half of timing word
    bool setTime              (uint64_t); /**< Set the time-stamp as a single 64-bit value */
    /// Set Hit probability of a pixel. Used when generating MicroSlices.
    bool setHitProbability    (float hit_prob);

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;              //!< Is the buffer internal?

    /// size of the MicroSlice _Buffer in words
    uint32_t sizeofdata() const;

    
  private:

    RawMicroSliceHeader    _Header;             //! MicroSlice Header
    RawTimingMarker        _TimingMarker;       //! MicroSlice Timing Marker
    mutable RawNanoSlice   _FloatingNanoSlice;  //! Floating NanoSlice
                                                //! One can set the pointer of the
                                                //! nanoslice to any one within the MicroSlice
    
    float                  _HitProbability;     //! Pixel Hit probability. Used when generating a MicroSlice


   //! Maximum size of the current buffer
    uint32_t               _max_buffer_size;

    /// The actual internal buffer (not a pointer)
    std::vector <uint32_t> _InternalBuffer;




  };
    
} // END NAMESPCE daqdataformats
#endif /* RAWMICROSLICEHEADER_H */
