#ifndef RAWRUN_H
#define RAWRUN_H

#include "DAQDataFormats/RawRunHeader.h"
#include "DAQDataFormats/RawRunTail.h"
#include "DAQDataFormats/RawConfigurationBlock.h"
#include "DAQDataFormats/RawEvent.h"


namespace daqdataformats {

    #define DEFAULT_RUN_SIZE 362144 /* Maximum Run Buffer Size  */

  class RawRun : public RawDAQData {
    
  public:
    // Constructors
    RawRun (uint32_t buffersize = DEFAULT_RUN_SIZE); //! The default constructor
    RawRun (const RawRun& copy_in);                  //! Deep copy constructor
    RawRun (void* millislice_buffer, uint32_t buffersize = DEFAULT_RUN_SIZE); //! Shallow copy constructor
    /// Destructor
    ~RawRun();

    /// Initialization of Data members
    void init();
    
    /// Deep Copy
    void copy(const RawRun& copy_in);

    /// Method to close the Run. Produces the tail and the CRC.
    bool close();
    

    //Add RawEvent into the buffer
    void addEvent(void* event_buffer);
    
    /// Read method is for reading the whole RawRun
    void* readData(void* run_buffer);
    // Suplementary reading methods
    void* readData(void* buffer, uint32_t size);
    void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);

    
    /// Size of Run _Buffer in words
    uint32_t sizeofdata() const;

    /// Print the Run structure
    void print() const;

    // Getters
    RawRunHeader*          getHeader();                  //! Run Header reference
    RawRunTail*            getTail();                    //! Run Tail reference
    RawConfigurationBlock* getConfigurationBlock();
    RawEvent*              getFloatingEvent();           //! The current floating Event reference

    /// Getters
    uint32_t getMarkerHi() const;
    uint32_t getMarkerLo() const;
    uint32_t getRunNumber() const;
    uint32_t getRunType() const;
    uint32_t getConfigIDX() const;
    uint32_t getPartition() const;
    uint32_t getSubrun() const;
    uint32_t getConfigurationBlockSet()  const;
    uint32_t getRunControlVersion() const;
    uint32_t getRunControlID() const;
    uint32_t getDataLoggerVersion() const;
    uint32_t getDataLoggerID() const;
    uint32_t getRunStartTime() const;
    uint32_t getSubRunStartTime() const;
    uint32_t getSubRunEndTime () const;
    uint32_t getRunEndTime() const;
    uint32_t getTotalEvents() const;
    uint32_t getUniversalEventStartLo() const;
    uint32_t getUniversalEventStartHi() const;
    uint64_t getUniversalEventStart()const;
    uint32_t getUniversalEventEndLo() const;
    uint32_t getUniversalEventEndHi() const;
    uint64_t getUniversalEventEnd() const;
    uint32_t getRunSize() const;
    uint32_t getTriggerVersion() const;
    uint32_t getTriggerCtrlID() const;
    uint32_t getTriggerMC() const;
    uint32_t getVTMod() const;
    uint32_t getValidTriggerTypesLow() const;
    uint32_t getValidTriggerTypesHigh() const;
    uint32_t getValidTriggerTypesHigh2() const;
    uint32_t getTriggerListIDX() const;
    uint32_t getTriggerPrescaleListIDX() const;
    uint32_t getFreeMemory() const;        //! Free memory left in the buffer.
    
    // Setters
    bool setConfigurationBlock    (void* configuration_block_pointer);
    bool setConfigurationBlock    (RawConfigurationBlock&  copyin);

    bool setFloatingEvent         (uint32_t ievent);     //! Set the Floating Event to the "ievent"-th Event in the Run
    bool setRunNumber             (uint32_t);
    bool setRunType               (uint32_t);
    bool setConfigIDX             (uint32_t);
    bool setPartition             (uint32_t);
    bool setSubrun                (uint32_t);
    bool setRunControlVersion     (uint32_t);
    bool setRunControlID          (uint32_t);
    bool setDataLoggerVersion     (uint32_t);
    bool setDataLoggerID          (uint32_t);
    bool setRunStartTime          (uint32_t);
    bool setSubRunStartTime       (uint32_t);
    bool setSubRunEndTime         (uint32_t);
    bool setRunEndTime            (uint32_t);
    bool setTriggerCtrlID         (uint32_t);
    bool setTriggerVersion        (uint32_t);
    bool setTriggerListIDX        (uint32_t);
    bool setTriggerPrescaleListIDX(uint32_t);





    
    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;        //!< Is the buffer internal?

  private:
    
    RawRunHeader           _Header;                  //! Run Header
    RawRunTail             _Tail;                    //! Run Tail
    RawConfigurationBlock  _ConfigurationBlock;      //! Configuration block

    RawEvent         _FloatingEvent;      //! Floating Event
    int              _FloatingEventNumber;//! Floating Event Number

   //! Maximum size of the current buffer
    uint32_t         _max_buffer_size;

    // The actual internal buffer (not a pointer).
    // Used only if the constructors with the default size are used
    std::vector <uint32_t>  _InternalBuffer;


  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWRUN_H */
