#ifndef DAQRAWTRIGGER_H
#define DAQRAWTRIGGER_H

#include <unistd.h>

#include "DAQDataFormats/TriggerDefines.h"

#include "DAQDataFormats/RawTriggerHeader.h"
#include "DAQDataFormats/RawTriggerTimingMarker.h"
#include "DAQDataFormats/RawTriggerRange.h"
#include "DAQDataFormats/RawTriggerTime.h"
#include "DAQDataFormats/RawTriggerMask.h"

#define COPY_COPIES_INTERNAL

namespace daqdataformats {

  //! Class to hold the data from the global trigger in the correct format
  /*!
    The RawTrigger class is designed to mimic the format of the actual
    data that emitted from the global trigger process.  This information
    include the source of the trigger, the trigger number etc...

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTrigger: public RawDAQData {
  
  public:
      
    RawTrigger();		                      //!< Constructor
    RawTrigger(void* input_buffer, uint32_t size=0);  //!< Constructor external buffer initialization
    RawTrigger(const RawTrigger& copy_in);	      //!< Copyin Constructor
    ~RawTrigger();                                    //!< Destructor
    
    bool copy(void* input_buffer, uint32_t size=0); //!< Copy method from an external data block (non-deep)
    bool copy(const RawTrigger& copy_in);           //!< Copy method from RawTrigger format (deep copy)

    // Get Methods for parts of the trigger block
    RawTriggerHeader*       getTriggerHeader()      {return &_theHeader;}
    RawTriggerTimingMarker* getTriggerTimingMarker(){return &_theMarker;}
    RawTriggerRange*        getTriggerRange()       {return &_theRange;}
    RawTriggerTime*         getTriggerTime()        {return &_theTime;}
    RawTriggerMask*         getTriggerMask()        {return &_theMask;}
    uint64_t                getTriggerStart()   {return _theMarker.getTriggerStart_baseClock();}

    // Set Methods for parts of the trigger block.  These are deep copy style set ops.
    bool setTriggerHeader      (RawTriggerHeader&       copyin){return _theHeader.copy(copyin);}
    bool setTriggerTimingMarker(RawTriggerTimingMarker& copyin){return _theMarker.copy(copyin);}
    bool setTriggerRange       (RawTriggerRange&        copyin){return _theRange.copy(copyin);}
    bool setTriggerTime        (RawTriggerTime&         copyin){return _theTime.copy(copyin);}
    bool setTriggerMask        (RawTriggerMask&         copyin){return _theMask.copy(copyin);}
    bool setTriggerStart       (uint64_t start_time)       {return _theMarker.setTriggerStart_baseClock(start_time);}
        
    bool reset();                                                  //!< Reset the class data
    bool clear();                                                  //!< Clear the buffer
    bool resetBufferSource();	                                   //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer);                      //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;

    //! General write method that writes the raw Trigger header formated data block to the target memory
    void*   writeData(void* buffer);			  //!< Write Data to target memory location
    ssize_t writeData(int fd);				  //!< Write Data to target file descriptor

    //! General read method that read the raw Trigger header formated data block from the target memory
    void*   readData(void* buffer);			  //!< Read Data from target memory location


    void print() const;


    //! General sizeof method which returns the size of the Trigger header formated data block in words (not bytes)
    uint32_t sizeofdata() const;	//!< Required sizeof method, returns the size of the object in 32bit words

    //! Assignment Operator
    bool operator=(const RawTrigger& copy_in);
    
  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Class initialization.  This method will zero associated buffers.
    
    //! The trigger header is a fixed size
    uint32_t BLOCK_WORDS;

    //! Maximum size of the current buffer (used for external buffers)
    uint32_t _max_buffer_size;

    //! Raw block structure containing the packed information
    //! This is allocated at initialization to become an array
    //! large enough to hold the data from all the resulting parts (header, Markers, etc...)
    uint32_t* _trigBuffer;

    //! Trigger Consists of the Following Items
    RawTriggerHeader       _theHeader;
    RawTriggerTimingMarker _theMarker;
    RawTriggerRange        _theRange;
    RawTriggerTime         _theTime;
    RawTriggerMask         _theMask;

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* DAQRAWTRIGGER_H */
