#ifndef DAQRAWTRIGGERMASK_H
#define DAQRAWTRIGGERMASK_H

#include <unistd.h>

#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/TriggerDefines.h"

#define COPY_COPIES_INTERNAL	/* Define the behavior of copy constructors */

namespace daqdataformats {
  
  enum TriggerMaskMASKS{
    MASK_MONTECARLO_FLAG   = 0x80000000,
    MASK_TRIG_MOD          = 0x3F000000,
    MASK_TRIG_TYPE_LOW     = 0x00FFFFFF,
    MASK_TRIG_TYPE_HIGH    = 0xFFFFFFFF,
    MASK_PRESCALE          = 0x7FFFFFFF,

    SHIFT_MONTECARLO_FLAG   = 31,
    SHIFT_TRIG_MOD          = 24,
    SHIFT_TRIG_TYPE_LOW     = 0,
    SHIFT_TRIG_TYPE_HIGH    = 0,
    SHIFT_PRESCALE          = 0,
  };

  

  //! Class to hold the data corresponding to the trigger mask
  /*!
    The RawTriggerMask contains the information on the type of trigger
    that is being issued.

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTriggerMask: public RawDAQData {
    
  public:
      
    RawTriggerMask();		                            //!< Constructor
    RawTriggerMask(void* input_buffer, uint32_t size = 0); //!< Constructor without data copy
    RawTriggerMask(const RawTriggerMask& copy_in);          //!< Copyin Constructor (deep copy)
    ~RawTriggerMask();                                      //!< Destructor
    
    bool copy(void* input_buffer, uint32_t size=0);       //!< Copy method from an external data block (non-deep)
    bool copy(const RawTriggerMask& copy_in);             //!< Copy method from RawTriggerMask format

    uint32_t getMonteCarloFlag() const; //! Get method for the Monte Carlo trigger flag
    uint32_t getTrigMod() const;        //! Get method for the trigger modification field
    uint32_t getTrigMask_Low() const;   //! Get method for the low trigger type bit mask (bits 0-31)
    uint32_t getTrigMask_High() const;  //! Get method for the high trigger type bit mask (bits 24- 55)
    uint32_t getTrigMask_High2() const; //! Get method for the hightest trigger type bit mask (56-88)
    uint32_t getTriggerMask() const;    //! Get just one number
    uint32_t getPrescale() const;       //! Get method for the trigger prescale

    bool setMonteCarloFlag(uint32_t state);    //!< Get method for the Monte Carlo trigger flag
    bool setTrigMod(uint32_t tmod);	           //!< Get method for the trigger modification field
    bool setTrigMask_Low(uint32_t mask);       //!< Get method for the low trigger type bit mask (bits 0-31)
    bool setTrigMask_High(uint32_t mask);      //!< Get method for the high trigger type bit mask (bits 24- 55)
    bool setTrigMask_High2(uint32_t mask);     //!< Get method for the hightest trigger type bit mask (56-88)
    bool setTriggerMask(uint32_t mask);        //!< set one of the 88 bits bits

    bool setPrescale(uint32_t prescale);       //!< Get method for the trigger prescale
            
    bool reset();                                         //!< Reset the class data
    bool clear();                                         //!< Clear the buffer
    bool resetBufferSource();	                          //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer);             //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;			  //!< Inquiry method for internal buffering type

    //! General write method that writes the raw Trigger Time formated data block to the target memory
    void* writeData(void* buffer);

    //! General read method that read the raw Trigger Time formated data block from the target memory
    void* readData(void* buffer);
    
    //! General sizeof method which returns the size of the Trigger Time formated data block in words (not bytes)
    uint32_t sizeofdata() const;

    void print() const;

    //! Assignment Operator
    bool operator=(const RawTriggerMask& copy_in);
    
  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Class initialization.  This method will zero associated buffers

    //! The trigger time is a fixed size
    const static uint32_t TRIGMASK_WORDS = 4;

    uint32_t _max_buffer_size;

    //! Pointer to internally managed data buffer
    uint32_t _trigMaskblock[TRIGMASK_WORDS];

  };
  
  
} // END NAMESPACE daqdataformats
#endif /* DAQRAWTRIGGERMASK_H */
