#ifndef DAQRAWTRIGGERTIMEMARKER_H
#define DAQRAWTRIGGERTIMEMARKER_H

#include <stdint.h>

#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/TriggerDefines.h"

#define COPY_COPIES_INTERNAL

namespace daqdataformats {
  
  //! Enumeration of the Masks and Shifts for the Trigger Timing Marker
  enum TriggerTimeMarkerMASKS{
    MASK_TRIGSTART_LOW_BASE_CLOCK  = 0xFFFFFFFF,
    MASK_TRIGSTART_HIGH_BASE_CLOCK = 0x00FFFFFF,
    MASK_TRIGSTART_LOW_833kHz      = 0xFFFFFFFF,
    MASK_TRIGSTART_HIGH_833kHz     = 0x00FFFFFF,

    SHIFT_TRIGSTART_LOW_BASE_CLOCK  = 0,
    SHIFT_TRIGSTART_HIGH_BASE_CLOCK = 0,
    SHIFT_TRIGSTART_LOW_833kHz      = 0,
    SHIFT_TRIGSTART_HIGH_833kHz     = 0,
  };

  //! Class that contains the timing information relating to a given trigger
  /*!
    The RawTriggerTimingMarker contains the start indicator for data collection.

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTriggerTimingMarker: public RawDAQData {
    
  public:
      
    RawTriggerTimingMarker();		                            //!< Constructor
    RawTriggerTimingMarker(void* input_buffer,uint32_t size=0);         //!< Constructor with data buffer initialization to external source and size
    RawTriggerTimingMarker(const RawTriggerTimingMarker& copy_in);  //!< Copyin Constructor
    ~RawTriggerTimingMarker();                                      //!< Destructor
    
    bool copy(void* input_buffer, uint32_t size=0);   //!< Copy method (non-deep) from an external buffer
    bool copy(const RawTriggerTimingMarker& copy_in);     //!< Copy method from RawTriggerTimingMarker format

    uint32_t getTriggerStartLow_baseClock() const;  //!< Get method for the low bits (0-31) of the trig start time on the base clock
    uint32_t getTriggerStartHigh_baseClock() const; //!< Get method for the high bits (32+) of the trig start time on the base clock
    uint64_t getTriggerStart_baseClock() const;     //!< Get method for the 64bit trig start time on the base clock (64 MHz, or 16MHz shifted by two bits)

    uint32_t getTriggerStartLow_833kHz() const; //! Get method for the low bits (0-31) of the trig start time on the 833kHz clock
    uint32_t getTriggerStartHigh_833kHz() const;//! Get method for the high bits (32+) of the trig start time on the 833kHz clock
    uint64_t getTriggerStart_833kHz() const;    //! Get method for the 64bit trig start time on the 833kHz clock

    bool setTriggerStartLow_baseClock(uint32_t start_low);      //! Set method for the low bits (0-31) of the trig start (Base Clock)
    bool setTriggerStartHigh_baseClock(uint32_t start_hi);      //! Set method for the high bits (32+) of the trig start (Base Clock)
    bool setTriggerStartLow_833kHz(uint32_t start_low);     //! Set method for the low bits (0-31) of the trig start (833kHz Clock)
    bool setTriggerStartHigh_833kHz(uint32_t start_hi);     //! Set method for the high bits (32+) of the trig start (833kHz Clock)
 
    bool setTriggerStart_baseClock(uint64_t start_time);     //! Set method for the 64bit timestamp on the base clock
    bool setTriggerStart_833kHz(uint64_t start_time);    //! Set method for the 64bit timestamp on the 833kHz clock
    bool setTimeTV(const struct timeval *);                            //! Set the timestamp as a struct timeval, sets both clocks timestamps

    bool reset();                             //!< General reset method
    bool clear();                             //!< Clear the buffer
    bool resetBufferSource();                 //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer); //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;			  //!< Inquiry method for internal buffering type

    //! General write method that writes the raw Trigger Time formated data block to the target memory
    void* writeData(void* buffer);

    //! General read method that read the raw Trigger Time formated data block from the target memory
    void* readData(void* buffer);
    
    //! General sizeof method which returns the size of the Trigger Time formated data block in words (not bytes)
    uint32_t sizeofdata() const;


    void print() const;
    
    bool operator=(const RawTriggerTimingMarker& copy_in);

  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Initialization

    //! The trigger time is a fixed size
    const static uint32_t TIMEMARKER_WORDS = 4;

    //! Maximum size of the current buffer (used for external buffers)
    uint32_t _max_buffer_size;

    //! Raw block structure containing the packed information
    uint32_t _trigTimeMarkerblock[TIMEMARKER_WORDS];
  };
  
  
} // END NAMESPCE daqdataformats
#endif /* DAQRAWTRIGGERTIMEMARKER_H */
