#include <sys/types.h>		// Standard System Types
#include <stdlib.h>		// Standard Library Header
#include <stdio.h>		// 
#include <time.h>		// Standard System Time Header

#include "DAQDataFormats/TimeStampCounter.h"	// High Resolution Time Stamp Counter

#include "DAQDataFormats/FEBSimulator.h"

#define DEFAULT_HIT_PROBABILITY 0.1
#define randReal() (1.0*rand() )/RAND_MAX

using namespace daqdataformats;

//------------------------Constructors------------------------------------------
//Default constructor. Use this causiously. FEBLinkID must be defined when generating a nanoslice.
FEBSimulator::FEBSimulator(){
  this->resetSEQ();
  this->setHitProbability(DEFAULT_HIT_PROBABILITY);	// Default hit probability
  fLastNanoSlice = new RawNanoSlice();
}

//------------------------------------------------------------------------------
// Constructor with assigning the FEBLink ID. Main constructor.
FEBSimulator::FEBSimulator(uint32_t linkID){
  this->resetSEQ();
  this->setLinkID(linkID);
  this->setHitProbability(DEFAULT_HIT_PROBABILITY);	// Default hit probability
  fLastNanoSlice = new RawNanoSlice();
}

//------------------------------------------------------------------------------
// Constructor with assigning the FEB LinkID and a probability. Use this one if you really going to simulate random hits
FEBSimulator::FEBSimulator(uint32_t linkID, float probability){
  this->resetSEQ();
  this->setLinkID(linkID);
  this->setHitProbability(probability);	// Default hit probability
  fLastNanoSlice = new RawNanoSlice();
}

//------------------------Destructor--------------------------------------------
FEBSimulator::~FEBSimulator() { delete fLastNanoSlice; }

//------------------------------------------------------------------------------
// Advance the FEB NanoSlice sequence number
void FEBSimulator::advanceSEQ(){
  fSEQ = (fSEQ++) & 0x7;		// The sequence number is only 3 bytes long
}

//------------------------------------------------------------------------------
// Reset the FEB NanoSlice sequence number
void FEBSimulator::resetSEQ(){
  fSEQ = 0x0;			// Reset the sequence number
}

//------------------------Setters-----------------------------------------------
void FEBSimulator::setLinkID(uint32_t linkID){
  fFEBLinkID = linkID;
}

//------------------------------------------------------------------------------
void FEBSimulator::setHitProbability(float probability){
  fHitProbability = probability;	// Set the prob of generating a single hit
}

//------------------NanoSlice Generators----------------------------------------
// Default generator. Randomly generates a NanoSlice on a random pixel with a current computer time for TimeStamp
// Might have an empty Nanoslice in the end.
// The output of this method is recorded in fLastNanoSlice.
void FEBSimulator::generateNanoSlice(){
  NanoSlice FullNanoSlice;

  // Fill the header information for the nano slice
  FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
  FullNanoSlice.Header.PixelAddress        = rand() & 0x001F;  // Generate a random Pixel
  FullNanoSlice.Header.FEBStatus           = 0x40;
  FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
  FullNanoSlice.Header.DebugMode           = false;
  FullNanoSlice.Header.DataPresent         = false;
  FullNanoSlice.Header.LinkStatus          = true;

  // Check to see if we should generate a hit for this block
  if( randReal() < fHitProbability){
    
    FullNanoSlice.Header.DataPresent = true;
    
    // Generate the time stamp off the internal clock
    cycle_t full_timestamp;

    full_timestamp.full_time = tsc();
    FullNanoSlice.TimeStamp = full_timestamp.time[0];

    // Generate the time stamp (just from the current time)
    // TimeStamp = time((time_t*)&TimeStamp);
    // TimeStamp = 0xFFFFFFFF;

    // Generate a random pulse height (12bit)
    FullNanoSlice.PulseHeight = rand() & 0x00000FFF;

    // Now Advance the Sequence Number
    this->advanceSEQ();
   }
  else{
    FullNanoSlice.TimeStamp = 0x0;
    FullNanoSlice.PulseHeight = 0x0;
  }

  //Now Set them parameters
  fLastNanoSlice->setParameters(FullNanoSlice);
}

//------------------------------------------------------------------------------
void FEBSimulator::generateNanoSlice(int PixelID){
NanoSlice FullNanoSlice;

  // Fill the header information for the nano slice
  FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
  FullNanoSlice.Header.PixelAddress        = PixelID & 0x001F;  // Pixel not random
  FullNanoSlice.Header.FEBStatus           = 0x40;
  FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
  FullNanoSlice.Header.DebugMode           = false;
  FullNanoSlice.Header.DataPresent         = false;
  FullNanoSlice.Header.LinkStatus          = true;

  // Check to see if we should generate a hit for this block
  if( randReal() < fHitProbability){

    FullNanoSlice.Header.DataPresent = true;

    // Generate the time stamp off the internal clock
    cycle_t full_timestamp;

    full_timestamp.full_time = tsc();
    FullNanoSlice.TimeStamp = full_timestamp.time[0];

    // Generate the time stamp (just from the current time)
    // TimeStamp = time((time_t*)&TimeStamp);
    // TimeStamp = 0xFFFFFFFF;

    // Generate a random pulse height (12bit)
    FullNanoSlice.PulseHeight = rand() & 0x00000FFF;

    // Now Advance the Sequence Number
    this->advanceSEQ();
   }
  else{
    FullNanoSlice.TimeStamp = 0x0;
    FullNanoSlice.PulseHeight = 0x0;
  }

  //Now Set them parameters
  fLastNanoSlice->setParameters(FullNanoSlice);

}

//------------------------------------------------------------------------------
// This version forces hit generation of a hit on a given pixel
void FEBSimulator::generateNanoSlice(int PixelID, bool ForceGeneration){
NanoSlice FullNanoSlice;

  // Fill the header information for the nano slice
  FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
  FullNanoSlice.Header.PixelAddress        = PixelID & 0x001F;
  FullNanoSlice.Header.FEBStatus           = 0x40;
  FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
  FullNanoSlice.Header.DebugMode           = false;
  FullNanoSlice.Header.DataPresent         = true;
  FullNanoSlice.Header.LinkStatus          = true;

  // Generate the time stamp off the internal clock
  cycle_t full_timestamp;

  full_timestamp.full_time = tsc();
  FullNanoSlice.TimeStamp = full_timestamp.time[0];

    // Generate the time stamp (just from the current time)
    // TimeStamp = time((time_t*)&TimeStamp);
    // TimeStamp = 0xFFFFFFFF;

  // Generate a random pulse height (12bit)
  FullNanoSlice.PulseHeight = rand() & 0x00000FFF;

  // Now Advance the Sequence Number
  this->advanceSEQ();
  //Now Set them parameters
  fLastNanoSlice->setParameters(FullNanoSlice);
}

// This version forces hit generation of a hit on a given pixel
void FEBSimulator::generateNanoSlice(int PixelID,int iTimeStamp, int iPulseHeight, bool ForceGeneration){

  NanoSlice FullNanoSlice;

  // Fill the header information for the nano slice
  FullNanoSlice.Header.FrameSequenceNumber = fSEQ;
  FullNanoSlice.Header.PixelAddress        = PixelID & 0x001F;
  FullNanoSlice.Header.FEBStatus           = 0x40;
  FullNanoSlice.Header.FEBLinkID           = fFEBLinkID;
  FullNanoSlice.Header.DebugMode           = false;
  FullNanoSlice.Header.DataPresent         = true;
  FullNanoSlice.Header.LinkStatus          = true;
  FullNanoSlice.TimeStamp                  = iTimeStamp;
  FullNanoSlice.PulseHeight                = iPulseHeight & 0x00000FFF;

  // Now Advance the Sequence Number
  this->advanceSEQ();
  // Now Set them parameters
  fLastNanoSlice->setParameters(FullNanoSlice);
}



  bool FEBSimulator::nonEmptyNanoSlice(){
  if(fLastNanoSlice->getHeader()->getDataPresent()){return true;}
  return false;
}


RawNanoSlice* FEBSimulator::getLastNanoSlice() {return fLastNanoSlice;}
