#include "DAQDataFormats/RawConfigurationBlock.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawConfigurationBlock::RawConfigurationBlock(uint32_t size){

    if (size < _Header.sizeofdata() + _Tail.sizeofdata()) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawConfigurationBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(_Header.sizeofdata() + _Tail.sizeofdata()) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }


    // Set the maximum size
    _max_buffer_size = size;

    _InternalBuffer.clear();
    _InternalBuffer.resize(_max_buffer_size, 0);

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}
//------------------------------------------------------------------------------
RawConfigurationBlock::RawConfigurationBlock(const RawConfigurationBlock& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawConfigurationBlock::RawConfigurationBlock(void* rawconfblock_buffer, uint32_t size){


    if (size < _Header.sizeofdata() + _Tail.sizeofdata()) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawConfigurationBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(_Header.sizeofdata() + _Tail.sizeofdata()) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // First, setup the internal buffer
        /// Set the maximum size
        _max_buffer_size = size;

        /// Allocate the internal Buffer
        _InternalBuffer.reserve(_max_buffer_size);
        /// Zero it out
        for(uint32_t i = 0; i < _max_buffer_size; ++i)
            _InternalBuffer[i] = 0;

        /// Set pointers to point to the internal buffer
        this->resetBufferSource();

        /// Initialize it
        this->init();


    this->setBufferSource(rawconfblock_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawConfigurationBlock::~RawConfigurationBlock() { }

// Initialize a new ConfigurationBlock (or reset it)
void RawConfigurationBlock::init(){

    /// Initialize the header
    _Header.init();
}

//------------------------------------------------------------------------------
void RawConfigurationBlock::copy(const RawConfigurationBlock& copy_in){

    // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    /// If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in.getBuffer());

    /// Copy the InternalBuffer for the Namespace, SystemID and Tail
    _Namespace.readData(copy_in._Namespace.getBuffer(), this->getNameSize());
    _SystemID.readData(copy_in._SystemID.getBuffer(), this->getSystemNumber()*SYSTEM_SIZE);
    _Tail.readData(copy_in._Tail.getBuffer());
}

void RawConfigurationBlock::addSystem(uint32_t sys_id, uint32_t sys_version){
    _SystemID.addSystem(sys_id, sys_version);
    _Header.advanceSystemNumber();
}

bool RawConfigurationBlock::close(){

    /// read the Namespace into the InternamBuffer right after the header
    this->readData(_Namespace.getBuffer(), _Namespace.sizeofdata(), _Header.sizeofdata());

    /// Now read the SystemID. pointer points to the Tail
    this->readData(_SystemID.getBuffer(), _SystemID.sizeofdata(), _Header.sizeofdata() + _Namespace.sizeofdata());

    /// Finally, read the tail
    this->readData(_Tail.getBuffer(), _Tail.sizeofdata(), _Header.sizeofdata() + _Namespace.sizeofdata() + _SystemID.sizeofdata());

    /// Set the Configuration Block size
    this->getHeader()->setConfigurationBlockSize(this->sizeofdata());

    /// The last thing is to add CRC
    this->addCheckSum();
    /// And copy it to the Tail
    _Tail.setCRC(RAWBUFF32[this->sizeofdata() - 1]);

    return true;
}


//------------------------------------------------------------------------------
// Getters
RawConfigurationHeader*    RawConfigurationBlock::getHeader()   { return &_Header; }   // Get ConfigurationBlock Header pointer
RawConfigurationTail*      RawConfigurationBlock::getTail()     { return &_Tail; }     // Get ConfigurationBlock Tail pointer
RawConfigurationNamespace* RawConfigurationBlock::getNamespace(){ return &_Namespace; }
RawConfigurationSystemID*  RawConfigurationBlock::getSystemID() { return &_SystemID; }
uint32_t RawConfigurationBlock::getMarkerHi1()                        const {return _Header.getMarkerHi1();}
uint32_t RawConfigurationBlock::getMarkerHi2()                        const {return _Header.getMarkerHi2();}
uint32_t RawConfigurationBlock::getMarkerLo1()                        const {return _Tail.getMarkerLo1();}
uint32_t RawConfigurationBlock::getMarkerLo2()                        const {return _Tail.getMarkerLo2();}
uint32_t RawConfigurationBlock::getConfigurationHeaderVersion()       const {return _Header.getConfigurationHeaderVersion();}
uint32_t RawConfigurationBlock::getDetectorID()                       const {return _Header.getDetectorID();}
uint32_t RawConfigurationBlock::getConfigurationBlockSize()           const {return _Header.getConfigurationBlockSize();}
uint32_t RawConfigurationBlock::getResourceManagerID()                const {return _Header.getResourceManagerID();}
uint32_t RawConfigurationBlock::getRunNumber()                        const {return _Header.getRunNumber();}
uint32_t RawConfigurationBlock::getRunType()                          const {return _Header.getRunType();}
uint32_t RawConfigurationBlock::getNameSize()                         const {return _Header.getNameSize();}
uint32_t RawConfigurationBlock::getPartition()                        const {return _Header.getPartition();}
uint32_t RawConfigurationBlock::getSubrun()                           const {return _Header.getSubrun();}
uint32_t RawConfigurationBlock::getSystemNumber()                     const {return _Header.getSystemNumber();}
uint32_t RawConfigurationBlock::getSystemVersionIDX()                 const {return _Header.getSystemVersionIDX();}
uint32_t RawConfigurationBlock::getDAQSystemVersion(uint32_t isystem) const {return _SystemID.getDAQSystemVersion(isystem);}
uint32_t RawConfigurationBlock::getDAQSystemID(uint32_t isystem)      const {return _SystemID.getDAQSystemID(isystem);}
uint32_t RawConfigurationBlock::getFreeMemory() const { return _max_buffer_size - this->sizeofdata();}
uint32_t RawConfigurationBlock::getNamespace( std::string& namespace_name){ return _Namespace.getNamespace(namespace_name);}


//------------------------------------------------------------------------------
//Setters
bool RawConfigurationBlock::setConfigurationHeaderVersion (uint32_t w) {return _Header.setConfigurationHeaderVersion(w);}
bool RawConfigurationBlock::setDetectorID                 (uint32_t w) {return _Header.setDetectorID(w);}
bool RawConfigurationBlock::setResourceManagerID          (uint32_t w) {return _Header.setResourceManagerID(w);}
bool RawConfigurationBlock::setRunNumber                  (uint32_t w) {return _Header.setRunNumber(w);}
bool RawConfigurationBlock::setRunType                    (uint32_t w) {return _Header.setRunType(w);}
bool RawConfigurationBlock::setPartition                  (uint32_t w) {return _Header.setPartition(w);}
bool RawConfigurationBlock::setSubrun                     (uint32_t w) {return _Header.setSubrun(w);}
bool RawConfigurationBlock::setSystemVersionIDX           (uint32_t w) {return _Header.setSystemVersionIDX(w);}
bool RawConfigurationBlock::setNamespace( std::string& namespace_name) {
    _Namespace.setNamespace(namespace_name);
    return _Header.setNameSize(_Namespace.sizeofdata());
}


bool RawConfigurationBlock::setBufferSource(void* pointer){
   _Buffer                = pointer;

   /// Set the Header's pointer
   _Header.setBufferSource(pointer);

   return true;
}


    bool RawConfigurationBlock::resetBufferSource(){
        return this->setBufferSource(&_InternalBuffer[0]);
    }

// clears all data in the buffer
    bool RawConfigurationBlock::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i < max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
    }
    bool RawConfigurationBlock::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawConfigurationBlock::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }


//------------------------------------------------------------------------------
// Print the ConfigurationBlock structure
void RawConfigurationBlock::print() const{
    std::cout<<"----------ConfigurationBlock----------------------------------------size="<<this->sizeofdata()<<std::endl;
    _Header.print();

    _Namespace.print();

    _SystemID.print();

    _Tail.print();

    
    /*
    std::cout<<"\nRAW:";
    for(int i = 0; i < this->sizeofdata(); ++i){
            printf(" ");
            this->printWord(i,false);
    }

    std::cout<<"\nRAW:";
    for(int i = 0; i < this->sizeofdata(); ++i){
            if(i > 0)std::cout<<"    ";
            this->printWord(i,true);
            std::cout<<"\n";
    }
*/

}

//------------------------------------------------------------------------------
// This is a general method to read the microslice from rawconfblock_buffer to _Buffer
void* RawConfigurationBlock::readData(void* rawconfblock_buffer){

    ///Read the header
    uint32_t* current_pointer = (uint32_t*)_Header.readData(rawconfblock_buffer);
    
    /// Read the rest of the ConfigurationBlock
    this->readData(current_pointer, this->getNameSize() + this->getSystemNumber()*SYSTEM_SIZE + _Tail.sizeofdata(), _Header.sizeofdata());


    /// re-read the namespace, systemid, and tail, since they are external
    current_pointer = (uint32_t*)_Namespace.readData(current_pointer, _Header.getNameSize());
    current_pointer = (uint32_t*)_SystemID.readData(current_pointer, _Header.getSystemNumber());
    current_pointer = (uint32_t*)_Tail.readData(current_pointer);

    return (void*)current_pointer;
}

//------------------------------------------------------------------------------
void* RawConfigurationBlock::readData(void* buffer_void, uint32_t size){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
// Method to read data from the buffer of size "size" into _Buffer starting from _Buffer_start_position position
void* RawConfigurationBlock::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){


  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
// ConfigurationBlock _Buffer size in words
uint32_t RawConfigurationBlock::sizeofdata()const {
  return _Header.sizeofdata() + _Namespace.sizeofdata() + _SystemID.sizeofdata() + _Tail.sizeofdata();
}
