#include "DAQDataFormats/RawMicroSlice.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawMicroSlice::RawMicroSlice(uint32_t size){

    if (size < RAWMICROSLICEHEADER_SIZE + RAWTIMINGMARKER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWMICROSLICEHEADER_SIZE + RAWTIMINGMARKER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }


    // Set the maximum size
    _max_buffer_size = size;


    /*
    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);
    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;
    */

    _InternalBuffer.clear();
    _InternalBuffer.resize(_max_buffer_size, 0);

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}
//------------------------------------------------------------------------------
RawMicroSlice::RawMicroSlice(const RawMicroSlice& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawMicroSlice::RawMicroSlice(void* microslice_buffer, uint32_t size){


    if (size < RAWMICROSLICEHEADER_SIZE + RAWTIMINGMARKER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWMICROSLICEHEADER_SIZE + RAWTIMINGMARKER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // First, setup the internal buffer
        // Set the maximum size
        _max_buffer_size = size;

        // Allocate the internal Buffer
        _InternalBuffer.reserve(_max_buffer_size);
        // Zero it out
        for(uint32_t i = 0; i < _max_buffer_size; ++i)
            _InternalBuffer[i] = 0;

        // Set pointers to point to the internal buffer
        this->resetBufferSource();

        // Initialize it
        this->init();


    this->setBufferSource(microslice_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawMicroSlice::~RawMicroSlice() { }

// Initialize a new MicroSlice (or reset it)
void RawMicroSlice::init(){

    // Initialize the header and timing marker
    _Header.init();
    _TimingMarker.init();

}

//------------------------------------------------------------------------------
void RawMicroSlice::copy(const RawMicroSlice& copy_in){

    // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in.getBuffer());
}

/*
bool RawMicroSlice::init(int rst_flag)
{
  // Reset Flag will cause the memory buffer to be
  // hard cleared to zero
  if(rst_flag & MICROSLICE_RESET_MEM){
    // Zero the Microslice Buffer
    for(int i=0; i< MAXIMUM_MICROSLICE_SIZE; ++i){
      microslice[i]=0;
    } // endfor i
    microslice_datasize =0;	// Zero the size of the data in the buffer
    buffer_current = microslice;	// Reset the insertion point
  } // endif rst_flag

  if(rst_flag & MICROSLICE_DELETE_FEB){
    for(int i=0; i< MAXIMUM_FRONTEND_BOARDS; ++i){
      if(FrontEndBoards[i]){
	delete FrontEndBoards[i];
      }	// endif
    } // endfor i
    Initialized_FEB=false;
  } // endif MICROSLICE_DELETE_FEB

  // Initialize a matrix of front end boards for generation
  if(rst_flag & MICROSLICE_INIT_FEB){
    for(int i=0; i< MAXIMUM_FRONTEND_BOARDS; ++i){
      FrontEndBoards[i] = new FEBSimulator(i);
    }
    Initialized_FEB=true;
  }
  
  // Reset the matrix of front end boards
  if(rst_flag & MICROSLICE_RESET_FEB){
    for(int i=0; i< MAXIMUM_FRONTEND_BOARDS; ++i){
      FrontEndBoards[i].resetSEQ();
      FrontEndBoards[i].setLinkID(i);
      FrontEndBoards[i].setHitProbability(0);
    }
    Initialized_FEB=true;
  }
    
  return true;
}
*/
 
//------------------------------------------------------------------------------
// Method to Add a packed Nanoslice to the _Buffer
// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
// and adds an appropriate flag to the MicroSlice Header.
void RawMicroSlice::addNanoSlice(void* buffer, bool check_mc_flag){
    this->addData(buffer, RAWNANOSLICE_SIZE, check_mc_flag);
}

//------------------------------------------------------------------------------
// Method to Add data to the _Buffer
// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
// and adds an appropriate flag to the MicroSlice Header.
void RawMicroSlice::addData(void* buffer, uint32_t nwords, bool check_mc_flag){

  if (this->getFreeMemory() < nwords && this->isInternalBuffer()) {
            std::string msg("Cannot add a NanoSlice. ");
            msg.append("MicroSlice buffer is full (total MicroSlice buffer size=");
            msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
            msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }


    // Add the Data into the _Buffer
  this->readData(buffer, nwords, this->sizeofdata());

 //Add byte count to the MicroSlice Header
  _Header.addByteCount(nwords * sizeof(uint32_t));



    // Check if the NanoSlice is the MC generated
    // If it is, set the appropriate MicroSlice MC flag
    if(!check_mc_flag)return;
    RawNanoSliceHeader nano_header(buffer);
    uint32_t FEBStatus = nano_header.getFEBStatus();
    if(FEBStatus > 0x0000003F) this->setMonteCarloFlag(true);
}


//------------------------------------------------------------------------------
// Method to generate a Microslice from scratch
void RawMicroSlice::generate(){

  // Set the MC flag
  this->setMonteCarloFlag(true);

  //Generate the timing marker of the microslice
  _TimingMarker.generate();

    // Loop over all the front end boards generating hits
    // and writing them to the buffers
    for(int boardID = 0; boardID < MAXIMUM_FRONTEND_BOARDS; ++boardID){
        //Declare the front end boards simulators
        FEBSimulator *FrontEndBoard = new FEBSimulator(boardID,_HitProbability);
        for(int pixelID = 0; pixelID < 32; ++pixelID){
                // Attempt to generate a nanoslice hit on a given pixel
                FrontEndBoard->generateNanoSlice(pixelID);
	
                // Check to see if it has nonzero data
                if(FrontEndBoard->nonEmptyNanoSlice()){
                    //Add the nanoslice
                    this->addNanoSlice(FrontEndBoard->getLastNanoSlice()->getBuffer());
                    } // endif NonEmptyNanoSlice
         } // endfor pixelID
        //Delete the front end boards simulator
        delete FrontEndBoard;
    } // endfor boardID
}// end of RawMicroSlice::unpack()


//------------------------------------------------------------------------------
// Getters
RawMicroSliceHeader*  RawMicroSlice::getHeader()               { return &_Header; }                    // Get MicroSlice Header pointer
RawTimingMarker*      RawMicroSlice::getTimingMarker()         { return &_TimingMarker; }              // Get Timing Marker pointer
RawNanoSlice*         RawMicroSlice::getFloatingNanoSlice()    { return &_FloatingNanoSlice;}
bool                  RawMicroSlice::getDataPresent()    const { return _Header.getDataPresent();}    // Get Method for Data present bit
bool                  RawMicroSlice::getMonteCarloFlag() const { return _Header.getMonteCarloFlag();} // Get Method for SIM bit
uint32_t              RawMicroSlice::getByteCount()      const { return _Header.getByteCount();}      // Get Method for Byte Count
uint32_t              RawMicroSlice::getLowWord()        const { return _TimingMarker.getLowWord();}  // Get method for lower half of timing word
uint32_t              RawMicroSlice::getHighWord()       const { return _TimingMarker.getHighWord();} // Get method for upper half of timing word
uint64_t              RawMicroSlice::getTime()           const { return _TimingMarker.getTime();}     // Get the time-stamp as a single 64-bit value */
float                 RawMicroSlice::getHitProbability() const { return _HitProbability;}
uint32_t              RawMicroSlice::getNumNanoSlices()  const { 
    return (this->sizeofdata() - _Header.sizeofdata() - _TimingMarker.sizeofdata()) / RAWNANOSLICE_SIZE;
}
uint32_t              RawMicroSlice::getFreeMemory()     const{
    if(this->sizeofdata() > _max_buffer_size)return 0;
    return _max_buffer_size - this->sizeofdata();
}
//------------------------------------------------------------------------------
//Setters
bool RawMicroSlice::setMonteCarloFlag (bool mc_flag)       { return _Header.setMonteCarloFlag(mc_flag);}
bool RawMicroSlice::setLowWord        (uint32_t word)      { return _TimingMarker.setLowWord (word);}
bool RawMicroSlice::setHighWord       (uint32_t word)      { return _TimingMarker.setHighWord(word);}
bool RawMicroSlice::setTime           (uint64_t timestamp) { return _TimingMarker.setTime(timestamp);}
bool RawMicroSlice::setHitProbability (float hit_prob) {
    _HitProbability = hit_prob;
    return true;
}
bool RawMicroSlice::setFloatingNanoSlice (uint32_t inano){

    if(inano >= this->getNumNanoSlices()){
        //std::cout<<"Warning from RawMicroSlice::setFloatingNanoSlice : setting floating nanoslice to "<<inano
        //         <<"-th nanoslice, whereas only "<<this->getNumNanoSlices()<<"nanoslices are in the current MicroSlice."<<"\n";
        return false;
    }


    uint32_t* pointer = &RAWBUFF32[0]
                      + _Header.sizeofdata()
                      + _TimingMarker.sizeofdata()
                      + inano * RAWNANOSLICE_SIZE;
    
    _FloatingNanoSlice.setBufferSource(pointer);
    return true;
}

bool RawMicroSlice::setBufferSource(void* pointer){
   _Buffer                = pointer;
   _Header.setBufferSource(pointer);
    uint32_t* pp = (uint32_t*)pointer;
    pp += _Header.sizeofdata();
    return _TimingMarker.setBufferSource(pp);
}
    bool RawMicroSlice::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawMicroSlice::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i < max_data; ++i)
            RAWBUFF32[i] = 0;

        return true;
    }
    bool RawMicroSlice::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawMicroSlice::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }


//------------------------------------------------------------------------------
// Print the MicroSlice structure
void RawMicroSlice::print() {
    std::cout<<"----------MicroSlice----------------------------------------size="<<this->sizeofdata()<<std::endl;
    _Header.print();
    _TimingMarker.print();

    /*
    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            printf(" ");
            this->printWord(i,false);
    }

    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            if(i > 0)std::cout<<"    ";
            this->printWord(i,true);
            std::cout<<"\n";
    }
    */

    for(uint32_t i = 0; i < this->getNumNanoSlices(); ++i){
        setFloatingNanoSlice(i);
        this->getFloatingNanoSlice()->print();
    }



}

//------------------------------------------------------------------------------
// MicroSlice _Buffer size in words
uint32_t RawMicroSlice::sizeofdata()const {
  return RAWMICROSLICEHEADER_SIZE + this->getByteCount()/4;
}


//------------------------------------------------------------------------------
// This is a general method to read the microslice from microslice_buffer to _Buffer
void* RawMicroSlice::readData(void* microslice_buffer){

    //Read the header and the timing marker
    microslice_buffer = _Header.readData(microslice_buffer);
    microslice_buffer = _TimingMarker.readData(microslice_buffer);

    uint32_t headerstuffsize = _Header.sizeofdata() + _TimingMarker.sizeofdata();

    //Now, read the rest of the the _Buffer
    return this->readData(microslice_buffer, this->sizeofdata() - headerstuffsize, headerstuffsize);
}

//------------------------------------------------------------------------------
void* RawMicroSlice::readData(void* buffer_void, uint32_t size){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
// Method to read data from the buffer of size "size" into _Buffer starting from _Buffer_start_position position
void* RawMicroSlice::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){


  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}
