#include "DAQDataFormats/RawMicroSliceHeader.h"
#include "DAQDataFormats/RawTimingMarker.h"
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif


using namespace daqdataformats;
using namespace rawmuhdr;
//------------------------------------------------------------------------------
// Default constructor
RawMicroSliceHeader::RawMicroSliceHeader(){

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    // Initialize it
    this->init();
}

//------------------------------------------------------------------------------
RawMicroSliceHeader::RawMicroSliceHeader(const RawMicroSliceHeader& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawMicroSliceHeader::RawMicroSliceHeader(void* microsliceheader_buffer){
    // First, initialize the Internal Buffer
            // Set pointers to point to the internal buffer
            this->resetBufferSource();

            // Initialize it
            this->init();

    // Now set the pointer to the external buffer
    this->setBufferSource(microsliceheader_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawMicroSliceHeader::~RawMicroSliceHeader() { }

//------------------------------------------------------------------------------
// Initialization
void RawMicroSliceHeader::init(){

    for(uint32_t i = 0; i < this->sizeofdata(); i++)
        RAWBUFF32[i] = 0;

    this->resetByteCount();
    
}

//------------------------------------------------------------------------------
// Deep Copy Method
void RawMicroSliceHeader::copy(const RawMicroSliceHeader& copy_in){

    // Copy the data from one buffer to the other
    for(int i = 0; i < RAWMICROSLICEHEADER_SIZE; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);
}

//------------------------------------------------------------------------------
// Getters
    bool      RawMicroSliceHeader::getDataPresent() const {
        if (BITFIELDGET(RAWBUFF32[0], MASK_DP , SHIFT_DP)) return true;
        return false;
    }
    bool      RawMicroSliceHeader::getMonteCarloFlag() const {
        if (BITFIELDGET(RAWBUFF32[0], MASK_SIM, SHIFT_SIM)) return true;
        return false;
    }
    uint32_t  RawMicroSliceHeader::getByteCount() const {
        return BITFIELDGET(RAWBUFF32[0], MASK_LeadingByteCount, SHIFT_LeadingByteCount);
    }

//------------------------------------------------------------------------------
// Setters
    bool RawMicroSliceHeader::setDataPresent(bool data_present){
        if(data_present)BITFIELDSET(RAWBUFF32[0], 1, MASK_DP , SHIFT_DP);
        else            BITFIELDSET(RAWBUFF32[0], 0, MASK_DP , SHIFT_DP);
        return true;
    }

    bool RawMicroSliceHeader::setMonteCarloFlag(bool mc_flag){
        if(mc_flag)     BITFIELDSET(RAWBUFF32[0], 1, MASK_SIM, SHIFT_SIM);
        else            BITFIELDSET(RAWBUFF32[0], 0, MASK_SIM, SHIFT_SIM);
        return true;
    }

    bool RawMicroSliceHeader::setByteCount(uint32_t LeadingByteCount){
        BITFIELDSET(RAWBUFF32[0], LeadingByteCount, MASK_LeadingByteCount, SHIFT_LeadingByteCount);
        return true;
    }

    bool RawMicroSliceHeader::setBufferSource(void* pointer){
        _Buffer = pointer;
        return true;
    }
    bool RawMicroSliceHeader::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawMicroSliceHeader::clear(){
        int max_data = this->sizeofdata();

        // Loop over the buffer and zero it
        for(int i=0; i< max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
    }
    bool RawMicroSliceHeader::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawMicroSliceHeader::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }

//------------------------------------------------------------------------------
// Add byte count. When adding a NanoSlice a ByteCount is increased by 12;
// After that. DataPresent is always true.
bool RawMicroSliceHeader::addByteCount(uint32_t nbytes){
    this->setByteCount(this->getByteCount() + nbytes);
    this->setDataPresent(true);
    return true;
}

//------------------------------------------------------------------------------
bool RawMicroSliceHeader::resetByteCount(){
    this->setByteCount(sizeof(uint32_t) * RAWTIMINGMARKER_SIZE);
    this->setDataPresent(false);
    return true;
}


//------------------------------------------------------------------------------
// Printing the microslice header
void RawMicroSliceHeader::print() const{
    printf("MicroSlice Header:");
    printf(" DP=%d", this->getDataPresent());
    printf(" MC=%d", this->getMonteCarloFlag());
    printf(" LeadingByteCount=%d",this->getByteCount());
 
    printf("      RAW=");
    this->printWord(0, false);
    printf("\nRAWHex\n");
    this->printWord(0, true);
}


//------------------------------------------------------------------------------
// Header size in words
    uint32_t  RawMicroSliceHeader::sizeofdata()const {return RAWMICROSLICEHEADER_SIZE;}
