#include "DAQDataFormats/RawMilliSliceIndex.h"
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif


using namespace daqdataformats;
using namespace rawmilliindx;

//------------------------------------------------------------
// Default constructor
RawMilliSliceIndex::RawMilliSliceIndex(uint32_t size){

    // Initialize index number to prevent errors (just in case)
    _IndexNumber = 0;

    // Set the maximum size
    _max_buffer_size = size;
    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);

    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;

    // Set pointers to point to the internal buffer
    this->resetBufferSource();
    
    this->init();
}

//------------------------------------------------------------------------------
RawMilliSliceIndex::RawMilliSliceIndex(const RawMilliSliceIndex& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawMilliSliceIndex::RawMilliSliceIndex(void* millisliceheader_buffer, uint32_t size, uint32_t indexnumber){

    // Set the maximum size
    _max_buffer_size = size;
    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);

    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;

    this->setBufferSource(millisliceheader_buffer);

    _IndexNumber = indexnumber;
}

//------------------------------------------------------------
// Destructor
RawMilliSliceIndex::~RawMilliSliceIndex() { }

//------------------------------------------------------------
// Initialize the data members
void RawMilliSliceIndex::init(){


    // Zero out the whole buffer
    for(uint32_t i = 0; i < this->sizeofdata(); i++)
        RAWBUFF32[i] = 0;

    _IndexNumber = 0;

}
//------------------------------------------------------------------------------
void RawMilliSliceIndex::copy(const RawMilliSliceIndex& copy_in){

    // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup other variables
    _IndexNumber             = copy_in._IndexNumber;

    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in.getBuffer());
}

//------------------------------------------------------------------------------
// Getters

    uint32_t RawMilliSliceIndex::getTime       (uint32_t iindex) const { return BITFIELDGET(RAWBUFF32[iindex*MILLISLICEINDEX_SIZE + WORD_Time]  , MASK_Time  , SHIFT_Time);}
    uint32_t RawMilliSliceIndex::getOffset     (uint32_t iindex) const { return BITFIELDGET(RAWBUFF32[iindex*MILLISLICEINDEX_SIZE + WORD_Offset], MASK_Offset, SHIFT_Offset);}
    uint32_t RawMilliSliceIndex::getIndexNumber()                const { return _IndexNumber;}

//------------------------------------------------------------------------------
// Add an index
bool RawMilliSliceIndex::addIndex (uint32_t time_length, uint32_t offset){
        this->setTime  (this->getIndexNumber(), time_length);
        this->setOffset(this->getIndexNumber(), offset);
        this->advanceIndexNumber();
        return true;
}
//------------------------------------------------------------------------------
// Setters
    // Set time of the "iindex"-th index
    bool RawMilliSliceIndex::setTime    (uint32_t iindex, uint32_t time_length){
        BITFIELDSET(RAWBUFF32[iindex*MILLISLICEINDEX_SIZE + WORD_Time]  , time_length  , MASK_Time  , SHIFT_Time);
        return true;
    }
    // Set Offset of the "iindex"-th index
    bool RawMilliSliceIndex::setOffset  (uint32_t iindex, uint32_t offset){
        BITFIELDSET(RAWBUFF32[iindex*MILLISLICEINDEX_SIZE + WORD_Offset], offset, MASK_Offset, SHIFT_Offset);
        return true;
    }
    bool RawMilliSliceIndex::setIndexNumber    (uint32_t indx){
        _IndexNumber = indx;
        return true;
    }
    bool RawMilliSliceIndex::advanceIndexNumber()             {
        _IndexNumber++;
        return true;
    }

    bool RawMilliSliceIndex::setBufferSource(void* pointer){
        if(pointer){
            this->_Buffer = pointer;
            return true;
        }
        return false;
    }
    bool RawMilliSliceIndex::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawMilliSliceIndex::clear(){
        int max_data = this->sizeofdata();

        // Loop over the buffer and zero it
        for(int i=0; i< max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
    }
    bool RawMilliSliceIndex::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawMilliSliceIndex::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }

    


    //------------------------------------------------------------------------------
// Print the MilliSlice header information
void RawMilliSliceIndex::print() const{
    printf("\nMilliSliceIndex:");

    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        printf(" ");
        this->printWord(i,false);
    }

    std::cout<<"\nMilliSliceIndex RAW bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        printf("\n");
        this->printWord(i,true);
    }

}

//------------------------------------------------------------------------------
// Size of _Buffer in words
uint32_t RawMilliSliceIndex::sizeofdata()const {return _IndexNumber * MILLISLICEINDEX_SIZE;}
