#include "DAQDataFormats/RawTriggerTimingMarker.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>
 
using namespace daqdataformats;

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
  RawTriggerTimingMarker::RawTriggerTimingMarker(){
    init();
  }
 
  RawTriggerTimingMarker::~RawTriggerTimingMarker(){
  }
 
 
  //
  // Constructor Methods
  //


RawTriggerTimingMarker::RawTriggerTimingMarker(void* input_buffer, uint32_t size){
  init(input_buffer, size);
}
  
// Deep Copy Constructor
RawTriggerTimingMarker::RawTriggerTimingMarker(const RawTriggerTimingMarker& copy_in){

#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    this->init(copy_in._Buffer, copy_in._max_buffer_size);
    _Buffer = ;
    
  }
#endif
  return;
}
 
// General Initialization 
void RawTriggerTimingMarker::init(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    this->setBufferSource(_trigTimeMarkerblock);
    this->clear();
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerTimingMarker constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TIMEMARKER_WORDS;
  }

  return;
}

// Copy from external source (non-deep copy)
bool RawTriggerTimingMarker::copy(void* input_buffer, uint32_t size){
  
  // Set the general buffer pointer to point at the internal buffer
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerTimingMarker copy method that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;    
  }else{
    _max_buffer_size = TIMEMARKER_WORDS;
  }

  return true;
}
 

// Deep Copy Method
bool RawTriggerTimingMarker::copy(const RawTriggerTimingMarker& copy_in){
  
  // Copy the raw packed block
  for(uint32_t i = 0; i < TIMEMARKER_WORDS; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor

  _max_buffer_size = copy_in._max_buffer_size;
  
  return true;
}
 
  //
  // GET METHODS
  //
 

  // get Method for low word of the time stamp
  uint32_t RawTriggerTimingMarker::getTriggerStartLow_baseClock() const {
    return BITFIELDGET(RAWBUFF32[0], MASK_TRIGSTART_LOW_BASE_CLOCK,
		       SHIFT_TRIGSTART_LOW_BASE_CLOCK);
  }

  // get Method for low word of the time stamp
  uint32_t RawTriggerTimingMarker::getTriggerStartHigh_baseClock() const {
    return BITFIELDGET(RAWBUFF32[1], MASK_TRIGSTART_HIGH_BASE_CLOCK,
		       SHIFT_TRIGSTART_HIGH_BASE_CLOCK);
  }

  // get Method for low word of the time stamp
  uint32_t RawTriggerTimingMarker::getTriggerStartLow_833kHz() const {
    return BITFIELDGET(RAWBUFF32[2], MASK_TRIGSTART_LOW_833kHz,
		       SHIFT_TRIGSTART_LOW_833kHz);
  }

  // get Method for low word of the time stamp
  uint32_t RawTriggerTimingMarker::getTriggerStartHigh_833kHz() const {
    return BITFIELDGET(RAWBUFF32[3], MASK_TRIGSTART_HIGH_833kHz,
		       SHIFT_TRIGSTART_HIGH_833kHz);
  }

  // get Method for 64 bit start time on the base clock
  uint64_t RawTriggerTimingMarker::getTriggerStart_baseClock() const {
      uint64_t timestamp;
      register uint32_t low, high; // hint to compiler to use registers

      low  = BITFIELDGET(RAWBUFF32[0], 
			 MASK_TRIGSTART_LOW_BASE_CLOCK, SHIFT_TRIGSTART_LOW_BASE_CLOCK);

      high = BITFIELDGET(RAWBUFF32[1],
			 MASK_TRIGSTART_HIGH_BASE_CLOCK, SHIFT_TRIGSTART_HIGH_BASE_CLOCK);

      timestamp=(uint64_t)high<<32 | low;
      return timestamp;
  }

  // get Method for 64 bit start time on the base clock
  uint64_t RawTriggerTimingMarker::getTriggerStart_833kHz() const {
      uint64_t timestamp;
      register uint32_t low, high;

      low  = BITFIELDGET(RAWBUFF32[2], 
			 MASK_TRIGSTART_LOW_833kHz, SHIFT_TRIGSTART_LOW_833kHz);
      high = BITFIELDGET(RAWBUFF32[3],
			 MASK_TRIGSTART_HIGH_833kHz, SHIFT_TRIGSTART_HIGH_833kHz);

      timestamp=(uint64_t)high<<32 | low;
      return timestamp;
  }


  //
  // SET METHODS
  //
 
  // set Method for low word of the time stamp
  bool RawTriggerTimingMarker::setTriggerStartLow_baseClock(uint32_t timestamp_low){
    BITFIELDSET( RAWBUFF32[0], timestamp_low, 
		 MASK_TRIGSTART_LOW_BASE_CLOCK, SHIFT_TRIGSTART_LOW_BASE_CLOCK);
    return true;
  }

  // set Method for low word of the time stamp
  bool RawTriggerTimingMarker::setTriggerStartHigh_baseClock(uint32_t timestamp_high){
    BITFIELDSET( RAWBUFF32[1], timestamp_high, 
		 MASK_TRIGSTART_HIGH_BASE_CLOCK, SHIFT_TRIGSTART_HIGH_BASE_CLOCK);
    return true;
  }

  // set Method for low word of the time stamp
  bool RawTriggerTimingMarker::setTriggerStartLow_833kHz(uint32_t timestamp_low){
    BITFIELDSET( RAWBUFF32[2], timestamp_low, 
		 MASK_TRIGSTART_LOW_833kHz, SHIFT_TRIGSTART_LOW_833kHz);
    return true;
  }

  // set Method for low word of the time stamp
  bool RawTriggerTimingMarker::setTriggerStartHigh_833kHz(uint32_t timestamp_high){
    BITFIELDSET( RAWBUFF32[3], timestamp_high, 
		 MASK_TRIGSTART_HIGH_833kHz, SHIFT_TRIGSTART_HIGH_833kHz);
    return true;
  }

  // set Method for 64bit time stamp
  bool RawTriggerTimingMarker::setTriggerStart_baseClock(uint64_t timestamp){
      register uint32_t low, high;

      low  = timestamp & 0xffffffff;
      high = timestamp >> 32;

      this->setTriggerStartHigh_baseClock(high);
      this->setTriggerStartLow_baseClock(low);
      return true;
  }

  // set Method for 64bit time stamp
  bool RawTriggerTimingMarker::setTriggerStart_833kHz(uint64_t timestamp){
      register uint32_t  low, high;

      low  = timestamp & 0xffffffff;
      high = timestamp >> 32;

      this->setTriggerStartHigh_833kHz(high);
      this->setTriggerStartLow_833kHz(low);
    
    return true;
  }

  // Set the buffer to use for storage.  This buffer can be external
  bool RawTriggerTimingMarker::reset(){
    this->resetBufferSource();
    this->clear();
    return true;
  }

  // clears all data in the buffer
  bool RawTriggerTimingMarker::clear(){
    int max_data = this->sizeofdata();

    // Loop over the buffer and zero it
    for(int i=0; i< max_data; ++i){
      RAWBUFF32[i]=0;
    }
    return true;
  }

  // Set the buffer to use for storage.  This buffer can be external
  bool RawTriggerTimingMarker::setBufferSource(void* input_buffer){
    this->_Buffer = input_buffer;
    return true;
  }

  // Set the buffer pointer to point back at the internal buffer
  bool RawTriggerTimingMarker::resetBufferSource(){
    this->_Buffer = _trigTimeMarkerblock; // set teh buffer pointer to point internally
    return true;
  }

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTriggerTimingMarker::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_trigTimeMarkerblock)){
    return true;
  }else{
    return false;
  }

  return false;
}

//
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTriggerTimingMarker::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){
    return 0;
  }

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTriggerTimingMarker::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){
    return 0;
  }

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTriggerTimingMarker::sizeofdata()const{
    return TIMEMARKER_WORDS;
}

bool RawTriggerTimingMarker::operator=(const RawTriggerTimingMarker& copy_in){
  return copy(copy_in); 
}

void RawTriggerTimingMarker::print() const{

    printf("++++ Trigger Timing Marker ++++\n");
    printf("Start Low Nova Clock:\t%u\n",this->getTriggerStartLow_baseClock() );
    printf("Start High Nova Clock:\t%u\n",this->getTriggerStartHigh_baseClock() );
    printf("Full Nova Clock:\t\t%Lu\n",(long long unsigned int)this->getTriggerStart_baseClock() );
    //printf("Start Time:\t\t%s\n",novadaq::timeutils::convertNovaTimeToString(this->getTriggerStart_baseClock()).c_str() );
    printf("Extraction Start Low:\t%u\n",this->getTriggerStartLow_833kHz() );
    printf("Extraction Start High:\t%u\n",this->getTriggerStartHigh_833kHz() );

}

